/*
This file is part of virtualGraffiti

virtualGraffiti is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

virtualGraffiti is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with virtualGraffiti,  If not, see <http://www.gnu.org/licenses/>.

Copyright 2009 Matthew Venn www.mattvenn.net
*/

import lll.wrj4P5.*;
import lll.Loc.*;
import processing.serial.*;
import ddf.minim.*;

Wrj4P5 wii;
/*

todo

*/

int savedImageCount = 0;

AudioSample rattle;
AudioPlayer spray;
Minim minim;

int correctX, correctY;
float a1,b1,c1,a3,b3,a2,b2,c2;

//the calibration pattern
PVector [] dots = new PVector[4];
//where we put the wiimote spots
PVector [] cal = new PVector[4];

boolean useWii = true;

int calPoints = 0;
boolean calibrated = false;
int calibrationDelay = 0;
boolean justCalibrated = false;
boolean lightDetected = false;
boolean waitingForNextCalibration = false;

int maxBrushSize = 70;
int minBrushSize = 20;
int wiiX;
int wiiY;

PFont fontA;
ColorPickerBox colorPicker;
Drips drips;
Can can;
Serial mySerial;
void setup() 
{
	size(1024,768,P2D);
	wii=new Wrj4P5(this);

	//dot is the original calibration image
	int calibInset = 20;
	dots[0] = new PVector( calibInset, calibInset ); //top left
	dots[1] = new PVector( width -calibInset, calibInset ); //top right
	dots[2] = new PVector( calibInset, height -calibInset ); //bot left
	dots[3] = new PVector( width -calibInset, height -calibInset); //bot right

	fontA = loadFont("Ziggurat-HTF-Black-32.vlw");
	textFont(fontA, 32);
        loadRandomBackground();
	
 
 //load the sounds 
        minim = new Minim(this);
  rattle = minim.loadSample("rattle.wav", 2048);
  if ( rattle == null ) println("Didn't get rattle!");
  spray = minim.loadFile("spraylong.wav", 2048);
  if( spray == null ) println( "didn't get spray!" );
 
 
	//setup color picker
	colorPicker = new ColorPickerBox( rattle, 5, 5, 100, 400, 12, 5 );

        //drips
        drips = new Drips( minBrushSize, maxBrushSize);
        
        //serial port to can
        String portName = Serial.list()[0];
        println( portName );
         mySerial = new Serial(this, portName, 57600);    

        can = new Can( spray, mySerial, minBrushSize, maxBrushSize );
       // can = new Can( this, minBrushSize, maxBrushSize );
      //  can.setupSerial();

  
	//connect wii
	if( useWii )
	{
		wii.connect(Wrj4P5.IR);
	}
}

 void serialEvent(Serial p) {
  can.updateCan2(mySerial.readString());
 }

 

 
void stop()
{
  // always close Minim audio classes when you are done with them
  rattle.close();
  spray.close();
  minim.stop();
 
}



void keyPressed()
{

  
   if( key == 'c' )
   {

  	if( calibrated == false )
	//load calibration from file
	{
		String[] lines;
		println( "loading calibration from file" );
		lines = loadStrings("calibration.txt");
		if( lines.length == 4 )
		{
			for( int i = 0; i < lines.length; i ++ )
			{
				println( i );
				String[] pieces = split(lines[i], '\t');
				println( "got peices" );
				if (pieces.length == 2) 
				{
					cal[i] = new PVector( int(pieces[0]), int(pieces[1]));
					println( cal[i].x + "," +  cal[i].y );
				}
			}
			if( calibrate() )	
			{
				calibrated = true;
				justCalibrated = true;
			}
		}
		else
		{
			println( "problem with calibration file..." );
		}
	}
	if( calibrated == true )
	{
		println( "storing calibration" );
		String[] lines = new String[4];
		for (int i = 0; i < 4; i++) 
		{
			lines[i] = cal[i].x + "\t" + cal[i].y;
		}
		saveStrings("calibration.txt", lines);
	}
   }
}

void draw()
{
	if( wii.isConnecting() )
	{
		background( 0 );
		fill( 255 );
		text("press 1&2 on wiimote...", 0, height /2 );
	}
	else
	{
		Loc p=wii.rimokon.irLights[0];
		if( useWii )
		{
			//spot found
			if (p.x>-1) 
			{
				lightDetected = true;
				//don't look for next cal straight away in case led switch causes flickers
				if( calibrationDelay -- == 0 ) waitingForNextCalibration = true;
				if( calibrationDelay < 0 ) calibrationDelay = 0;
				wiiX = (int)( ( 1 - p.x) * width );
				wiiY = (int)( ( 1 - p.y) * height );
			}
			else
			{
				calibrationDelay = 10;
				lightDetected = false;
			}
		}

		if( useWii && calibrated == false )
		{
			if( waitingForNextCalibration == true && calibrationDelay == 0 )
			{
				getCalibrationPoint();
				//wait for light to go on again
				calibrationDelay = 10;
				waitingForNextCalibration = false;
				delay( 500 );
			}

			drawCalibrationImage();  
			fill( 255 );
			noStroke();
			if( calibrationDelay == 0 ) fill( color( 255, 0, 0 ) );
			
			ellipse( wiiX, wiiY, 20,20 );
			text("calibration... " + calibrationDelay, 0, height /2 - 40 );

			//previous points
			for( int i = 0; i < calPoints ; i ++ )
			{
				text( i + " : " + cal[i].x + " , ", 0, height / 2 + 40 * i);
				text(cal[i].y, 200, height / 2 + 40 * i);
			}
			//current point
			text( calPoints + " : " + wiiX + " , ", 0, height / 2 + 40 * calPoints);
				text(wiiY, 200, height / 2 + 40 * calPoints);

		}
		else if( useWii && calibrated == false && justCalibrated == true )
		{
			//just failed to calibrate
			delay( 1000 );
		}
		else if( useWii && calibrated == true && justCalibrated == true )
		{
			background( 0 );
			fill( 255 );
			text("calibration OK", 0, height /2 - 40 );
			justCalibrated = false;
			delay( 500 );
			//wipe it
                        loadRandomBackground();
		}
		//it's paintin' time!
		else
		{
                        //read new info from the can 
                       // can.updateCan();
                         
                        if( colorPicker.getWipeButton() )
			{
				loadRandomBackground();

			}
			colorPicker.display();
			noStroke();
			if( useWii )
			{
				if(p.x>-1)
                                {
					translateCoords( wiiX, wiiY );
				
					if( ! spray.isPlaying() ) spray.loop();


					//int size = (int)(p.z * 100);
					int size = can.getBrushSize(); //colorPicker.getBrushSize();
					drips.addMouseHistory( correctX, correctY );
					drips.addDrip( correctX, correctY, size, colorPicker.getCurrentColor() ); 
colorPicker.update( correctX, correctY );
					fill( colorPicker.getCurrentColor(), can.getOpacity() );					
ellipse( correctX, correctY, size, size );
					//println( size );
				}
else
{
      if( spray.isPlaying() ) spray.pause();
}
                                 
			}
			else
			{
				if( mousePressed ) 
				{
					drips.addMouseHistory( mouseX, mouseY );
					int size = can.getBrushSize(); //colorPicker.getBrushSize();

					drips.addDrip( mouseX, mouseY, size, colorPicker.getCurrentColor( ) ); 
					colorPicker.update( mouseX, mouseY );
					fill( colorPicker.getCurrentColor(), can.getOpacity() );
					ellipse( mouseX, mouseY, size, size );
				}
			}
			//draw the drips
			drips.drawDrips();
//println( can.getBrushSize() );
		}
	}
}

void drawCalibrationImage()
{
	int lineLength = 10;

	background( 0 );
	stroke( 255 );
	fill( 0 );
	ellipse( dots[calPoints].x, dots[calPoints].y, lineLength * 2, lineLength * 2 );
	line( 	dots[calPoints].x - lineLength, dots[calPoints].y,
		dots[calPoints].x + lineLength, dots[calPoints].y );
	line( 	dots[calPoints].x, dots[calPoints].y - lineLength,
		dots[calPoints].x, dots[calPoints].y + lineLength );

	//draw the points already got
	fill( 255 );
	for( int i = 0; i < calPoints ; i ++ )
	{
		ellipse( cal[i].x, cal[i].y, 20, 20 );
	}

}

void translateCoords(int X, int Y )
{
	float corrX = (a1 * X + b1 * Y + c1 ) / (a3 * X + b3 * Y + 1 );
	float corrY = (a2 * X + b2 * Y + c2 ) / (a3 * X + b3 * Y + 1 );

	correctX = (int)corrX;
	correctY = (int)corrY;
	//println( (int)corrX + ":" + (int)corrY );
}


void getCalibrationPoint()
{
//	if( calibrated == false )
	{
		cal[calPoints ++] = new PVector( wiiX, wiiY );
		//println( cal[calPoints].x + " " + cal[calPoints].y );
		//wait so we don't get triggered again immediately
//		delay( 500 );

		if( calPoints == 4 )
		{
			if( calibrate() )
			{
				calibrated = true; 
				justCalibrated = true;
				calPoints = 0;
			}
			else
			{
				//recalibrate
				justCalibrated = true;
				calPoints = 0;
			}
			
		}
	}

}


void loadRandomBackground()
{
  savedImageCount ++;
  String saveName = "image" + savedImageCount + ".jpg";
  save( saveName );
  String path = sketchPath + "/data/";
  File file = new File(path);
  ArrayList images = new ArrayList();
    File[] files = file.listFiles();
 println( "looking in" + path );
  for (int i = 0; i < files.length; i++)
  {
    File f = files[i];    
    String fileName = f.getName();
    if( fileName.endsWith( ".jpg" ) )
    {
     images.add( fileName );
     
    }
  }
  int randomImageNumber = (int)random( 0, images.size() );

  PImage bg = loadImage( (String)images.get( randomImageNumber ) );
  background(bg);
 
}

boolean calibrate()
{
	println( "running calibration" );

	float [][] matrix = { 
		{ -1, -1, -1, -1, 0, 0, 0, 0 },
		{   -cal[0].x, -cal[1].x, -cal[2].x, -cal[3].x, 0, 0, 0, 0 },
		{ -cal[0].y, -cal[1].y, -cal[2].y, -cal[3].y, 0,0,0,0 },
		{ 0,0,0,0,-1,-1,-1,-1 },
		{ 0,0,0,0, -cal[0].x, -cal[1].x, -cal[2].x, -cal[3].x },
		{ 0,0,0,0, -cal[0].y, -cal[1].y, -cal[2].y, -cal[3].y },
		{ cal[0].x * dots[0].x, cal[1].x * dots[1].x, cal[2].x * dots[2].x, cal[3].x * dots[3].x, cal[0].x * dots[0].y, cal[1].x * dots[1].y, cal[2].x * dots[2].y, cal[3].x * dots[3].y },
		{ cal[0].y * dots[0].x, cal[1].y * dots[1].x, cal[2].y * dots[2].x, cal[3].y * dots[3].x, cal[0].y * dots[0].y, cal[1].y * dots[1].y, cal[2].y * dots[2].y, cal[3].y * dots[3].y },
		};

	float [] bb = { -dots[0].x, -dots[1].x, -dots[2].x, -dots[3].x, -dots[0].y, -dots[1].y, -dots[2].y, -dots[3].y };

	// gauß-elimination
	for( int j = 1; j < 4; j ++ )
	{
		for( int i = 1; i < 8; i ++ )
		{
			matrix[i][j] = - matrix[i][j] + matrix[i][0];
		}
		bb[j] = -bb[j] + bb[0];
		matrix[0][j] = 0;
	}


	for( int i = 2; i < 8; i ++ )
	{
		matrix[i][2] = -matrix[i][2] / matrix[1][2] * matrix[1][1] + matrix[i][1];
	}
	bb[2] = - bb[2] / matrix[1][2] * matrix[1][1] + bb[1];
	matrix[1][2] = 0;


	for( int i = 2; i < 8; i ++ )
	{
		matrix[i][3] = -matrix[i][3] / matrix[1][3] * matrix[1][1] + matrix[i][1];
	}
	bb[3] = - bb[3] / matrix[1][3] * matrix[1][1] + bb[1];
	matrix[1][3] = 0;



	for( int i = 3; i < 8; i ++ )
	{
		matrix[i][3] = -matrix[i][3] / matrix[2][3] * matrix[2][2] + matrix[i][2];
	}
	bb[3] = - bb[3] / matrix[2][3] * matrix[2][2] + bb[2];
	matrix[2][3] = 0;

	for( int j = 5; j < 8; j ++ )
	{
		for( int i = 4; i < 8; i ++ )
		{
			matrix[i][j] = -matrix[i][j] + matrix[i][4];
		}
		bb[j] = -bb[j] + bb[4];
		matrix[3][j] = 0;
	}

	for( int i = 5; i < 8; i ++ )
	{
		matrix[i][6] = -matrix[i][6] / matrix[4][6] * matrix[4][5] + matrix[i][5];
	}

	bb[6] = - bb[6] / matrix[4][6] * matrix[4][5] + bb[5];
	matrix[4][6] = 0;


	for( int i = 5; i < 8; i ++ )
	{
		matrix[i][7] = -matrix[i][7] / matrix[4][7] * matrix[4][5] + matrix[i][5];
	}
	bb[7] = - bb[7] / matrix[4][7] * matrix[4][5] + bb[5];
	matrix[4][7] = 0;


	for( int i = 6; i < 8; i ++ )
	{
		matrix[i][7] = -matrix[i][7] / matrix[5][7] * matrix[5][6] + matrix[i][6];
	}
	bb[7] = - bb[7] / matrix[5][7] * matrix[5][6] + bb[6];
	matrix[5][7] = 0;

	matrix[7][7] = - matrix[7][7]/matrix[6][7]*matrix[6][3] + matrix[7][3];
	bb[7] = -bb[7]/matrix[6][7]*matrix[6][3] + bb[3];
	matrix[6][7] = 0;

	/*
	println( "data dump" );
	for( int i = 0; i < 8 ; i ++ )
	{
	for( int j= 0; j < 8 ; j ++ )
	{
	print( matrix[i][j] + "," );
	}
	println("");
	}

	println( "bb" );
	for( int j= 0; j < 8 ; j ++ )
	{
	print( bb[j] + "," );
	}

	println("");
	*/

	b3 =  bb[7] /matrix[7][7];
	a3 = (bb[3]-(matrix[7][3]*b3))/matrix[6][3];
	b2 = (bb[6]-(matrix[7][6]*b3+matrix[6][6]*a3))/matrix[5][6];
	a2 = (bb[5]-(matrix[7][5]*b3+matrix[6][5]*a3+matrix[5][5]*b2))/matrix[4][5];
	c2 = (bb[4]-(matrix[7][4]*b3+matrix[6][5]*a3+matrix[5][4]*b2+matrix[4][4]*a2))/matrix[3][4];
	b1 = (bb[2]-(matrix[7][2]*b3+matrix[6][2]*a3+matrix[5][2]*b2+matrix[4][2]*a2+matrix[3][2]*c2))/matrix[2][2];
	a1 = (bb[1]-(matrix[7][1]*b3+matrix[6][1]*a3+matrix[5][1]*b2+matrix[4][1]*a2+matrix[3][1]*c2+matrix[2][1]*b1))/matrix[1][1];
	c1 = (bb[0]-(matrix[7][0]*b3+matrix[6][0]*a3+matrix[5][0]*b2+matrix[4][0]*a2+matrix[3][0]*c2+matrix[2][0]*b1+matrix[1][0]*a1))/matrix[0][0];

	if( Float.isNaN( b3 ) ) return false;
	if( Float.isNaN( b2 ) ) return false;
	if( Float.isNaN( a2 ) ) return false;
	if( Float.isNaN( c2 ) ) return false;
	if( Float.isNaN( a3 ) ) return false;
	if( Float.isNaN( b1 ) ) return false;
	if( Float.isNaN( a1 ) ) return false;
	if( Float.isNaN( c1 ) ) return false;
	println( "calibrated OK" );
	return true;
}
